(* GNU-GPL
 * System Compression - Logiciel de Compression/Decompression
 * Copyright (c) 2002 Ulrich Van Den Hekke (ulrich.vdh@free.fr)
 * 24 rue Jean Jaures
 * 59560 Comines
 *
 * Ce programme est libre, vous pouvez le redistribuer et/ou le modifier selon
 * les termes de la Licence Publique Gnrale GNU publie par la
 * Free Software Foundation (version 2 ou bien toute autre version ultrieure
 * choisie par vous).
 *
 * Ce programme est distribu car potentiellement utile, mais
 * SANS AUCUNE GARANTIE, ni explicite ni implicite, y compris les garanties de
 * commercialisation ou d'adaptation dans un but spcifique. Reportez-vous  la
 * Licence Publique Gnrale GNU pour plus de dtails.
 *
 * Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en mme
 * temps que ce programme ; si ce n'est pas le cas, crivez  la
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, tats-Unis.
 *
 *)

(******************************************************************************)
(* Unit contenant des classes d'encapsulation des plugins                    *)
(* Version : 1.0.0.0                                                          *)
(* Author : Ulrich Van Den Hekke                                              *)
(* Create day : 06/07/2002                                                    *)
(******************************************************************************)
unit AlgoList;

interface

uses Windows, Classes, SysUtils, Common, IniFiles;

type
  {$IFDEF CONSOLEAPP}
  TOnProgression = function(const Algo: string; const Position, Max: LongWord): boolean;
  {$ELSE}
  TOnProgression = function(const Algo: string; const Position, Max: LongWord): boolean of object;
  {$ENDIF}

  TAlgoPlugins = class
  private
    // Information du module
    HModule      : HINST;
    FFileName    : TFileName;

    // Adresse des procdure de la librairie
    FInfoProc     : TInfoProc;
    FAboutProc    : TAboutProc;
    FPackProc     : TPackProc;
    FUnpackProc   : TUnpackProc;
    FOnProgression: TOnProgression;

    // TStream d'entre et de sortie
    FInputStream,
    FOutputStream : TStream;

    // Buffer de lecture et d'criture
    FInputBuffer,
    FOutputBuffer : array of Byte;

    // Position dans le buffer et Taille
    FInputPos,
    FOutputPos,
    FInputMax,
    FOutputMax    : Integer;

    // Position total dans le fichier
    FCountSize,
    FInputPosFile : LongWord;

    function GetPluginsInformation: TInformations;

    procedure Init(Count: Integer; Input, Output: TStream);
    procedure Uninit();
  protected
    function  GetByte (var Buffer: Byte): boolean; stdcall;
    procedure PutByte (const Buffer: Byte; const flush: boolean); stdcall;

    function PackTest(St: TStream; const count: integer): Integer;
  public
    constructor Create(FileName: TFileName);
    destructor Destroy(); override;

    procedure About(hWnd: HWND);

    procedure Compression(var ByteWritten: Integer; const Count: Integer; const Input, Output: TStream);
    procedure Decompression(const TextSize, Count: Integer; const Input, Output: TStream);

    // function de test
    function RandomStreamTest(Size: Integer): TStream;
    function PackStreamTest(St: TStream; const count: integer = -1): TStream;
    function UnpackStreamTest(OriginalSize: Integer; St: TStream; const count: integer = -1): TStream;
    function CompareStreamTest(S1, S2: TStream): boolean;

    property PluginsInformation: TInformations  read GetPluginsInformation;
    property FileName          : TFileName      read FFileName;
    property OnProgression     : TOnProgression read FOnProgression write FOnProgression;
  end;

  TAlgoPluginsList = class
  private
    FIniFile: TIniFile;
    FDirName: TFileName;
    FList   : TList;
    FOnProgression: TOnProgression;
    function GetItems(Index: Integer): TAlgoPlugins;
    function GetCount: Integer;
  protected
  public
    constructor Create(DirName: TFileName);
    destructor Destroy(); override;

    function IndexOf(S: String): Integer; overload;
    function IndexOf(S: TInfosSerial): Integer; overload;
    function IndexOfBest(const St: TStream; const Count: integer): integer;

    procedure AddPlugins(FileName: TFileName);
    procedure RemovePlugins(Index: Integer);

    property Items[Index: Integer]: TAlgoPlugins read GetItems; default;
    property Count: Integer read GetCount;

    property OnProgression: TOnProgression read FOnProgression write FOnProgression;
  end;

implementation

uses FileTools;

resourcestring
  rsEOutOfMemory      = 'Il n''y a pas suffisamment de mmoire pour (De)Compresser.'#10 +
                        'Le logiciel a besoin d''un minimum de %d octets';
  rsRemPluginsFile    = '%s\remalg.ini';
  rsRemPluginsSection = 'Rem Plugins';
  rsDateTime          = 'Last Date Time';

const
  BufferLength = 64 * 1024; { 64 Kb }

type
  EMemory = class(Exception);

{ TAlgoPlugins }

procedure TAlgoPlugins.About(hWnd: HWND);
begin
  FAboutProc(hWnd);
end;

procedure TAlgoPlugins.Compression(var ByteWritten: Integer; const Count: Integer; const Input, Output: TStream);
begin
  try
    Init(Count, Input, Output);
  except
    raise EMemory.CreateFmt(rsEOutOfMemory, [2 * BufferLength]);
  end;

  try
    FPackProc(ByteWritten, GetByte, PutByte);
  finally
    Uninit();
  end;
end;

constructor TAlgoPlugins.Create(FileName: TFileName);
begin
  FFileName := FileName;
  HModule   := LoadLibrary(PChar(FFileName));

  FInfoProc    := GetProcAddress(HModule, 'Informations');
  FAboutProc   := GetProcAddress(HModule, 'About');
  FPackProc    := GetProcAddress(HModule, 'Pack');
  FUnpackProc  := GetProcAddress(HModule, 'Unpack');
end;

procedure TAlgoPlugins.Decompression(const TextSize, Count: Integer; const Input, Output: TStream);
begin
  try
    Init(Count, Input, Output);
  except
    raise EMemory.CreateFmt(rsEOutOfMemory, [2 * BufferLength]);
  end;

  try
    FUnpackProc(TextSize, GetByte, PutByte);
  finally
    Uninit();
  end;
end;

destructor TAlgoPlugins.Destroy;
begin
  FreeLibrary(HModule);
end;

procedure TAlgoPlugins.Uninit;
begin
  FInputStream  := nil;
  FOutputStream := nil;

  SetLength(FInputBuffer, 0);
  SetLength(FOutputBuffer, 0);
end;

function TAlgoPlugins.GetByte(var Buffer: Byte): boolean;
begin
  result := (not ((FInputMax <> BufferLength) and (FInputPos = FInputMax))) and (FInputPosFile < FCountSize);
  if(result = false) then Exit;

  if(FInputPos >= BufferLength) then
  begin
    if(FCountSize - FInputPosFile < BufferLength) then
      FInputMax := FInputStream.Read(FInputBuffer[0], FCountSize - FInputPosFile)
    else
      FInputMax := FInputStream.Read(FInputBuffer[0], BufferLength);
    FInputPos := 0;
  end;

  if(FInputMax <> 0) then
  begin
    Buffer := FInputBuffer[FInputPos];
    Inc(FInputPos);
    Inc(FInputPosFile);
    if(Assigned(OnProgression)) then result := OnProgression(PluginsInformation.iName, FInputPosFile, FCountSize);
  end else result := false;
end;

function TAlgoPlugins.GetPluginsInformation: TInformations;
begin
  result := FInfoProc;
end;

procedure TAlgoPlugins.Init(Count: Integer; Input, Output: TStream);
begin
  SetLength(FInputBuffer, BufferLength);
  SetLength(FOutputBuffer, BufferLength);

  FInputPos     := BufferLength;
  FOutputPos    := 0;

  FInputMax     := BufferLength;
  FOutputMax    := BufferLength;

  FInputStream  := Input;
  FOutputStream := Output;

  FCountSize    := Count;
  FInputPosFile := 0;
end;

procedure TAlgoPlugins.PutByte(const Buffer: Byte; const flush: boolean);
begin
  if((FOutputPos >= BufferLength) or flush) then begin
    FOutputStream.Write(FOutputBuffer[0], FOutputPos);
    FOutputPos := 0;
  end;

  if(not flush) then begin
    FOutputBuffer[FOutputPos] := Buffer;
    Inc(FOutputPos);
  end;
end;

function TAlgoPlugins.CompareStreamTest(S1, S2: TStream): boolean;
var I     : Integer;
    B1, B2: Byte;
begin
  S1.Seek(0, soFromBeginning); S2.Seek(0, soFromBeginning);
  result := true;
  if(S1.Size <> S2.Size) then begin
    Result := false;
    Exit;
  end;
  for I := 1 to S1.Size do begin
    S1.Read(B1, 1); S2.Read(B2, 1);
    if(B1 <> B2) then begin
      result := false;
      Exit;
    end;
    if(Assigned(OnProgression)) then OnProgression(PluginsInformation.iName, I, S1.Size);
  end;
end;

function TAlgoPlugins.PackStreamTest(St: TStream; const count: integer): TStream;
var Cnt: integer;
begin
  Result        := TMemoryStream.Create;
  try
    if(Count = -1) then Cnt := St.Size else Cnt := Count;
    St.Seek(0, soFromBeginning);
    Compression(Cnt, Cnt, St, Result);
  except
    Result.Free;
    Result := nil;
  end;
end;

function TAlgoPlugins.RandomStreamTest(Size: Integer): TStream;
var I: Integer;
    B: Byte;
begin
  Result        := TMemoryStream.Create;
  try
    Result.Size := Size;

    Randomize();
    for I := 1 to Size do begin
      B := Random(255);
      Result.Write(B, SizeOf(Byte));
      if(Assigned(OnProgression)) then OnProgression(PluginsInformation.iName, I, Size);
    end;
  except
    Result.Free;
    Result := nil;
  end;
end;

function TAlgoPlugins.UnpackStreamTest(OriginalSize: Integer; St: TStream; const count: integer): TStream;
var Cnt: integer;
begin
  Result        := TMemoryStream.Create;
  try
    if(Count = -1) then Cnt := St.Size else Cnt := Count;
    St.Seek(0, soFromBeginning);
    Decompression(OriginalSize, Cnt, St, Result);
  except
    Result.Free;
    Result := nil;
  end;
end;

function TAlgoPlugins.PackTest(St: TStream; const count: integer): Integer;
var S: TStream;
begin
  S := nil;
  try
    try
      S := PackStreamTest(St, count);
      result := S.Size;
    except
      result := 0;
      raise;
    end;
  finally
    S.Free;
  end;
end;

{ TAlgoPluginsList }

procedure TAlgoPluginsList.AddPlugins(FileName: TFileName);
begin
  FList.Add(TAlgoPlugins.Create(FDirName + FileName));
  FIniFile.WriteBool(rsRemPluginsSection, ExtractFileName(FileName), false);
end;

constructor TAlgoPluginsList.Create(DirName: TFileName);
var SRec: TSearchRec;
    Res : Integer;
begin
  FDirName       := NormalDir(DirName);
  FIniFile       := TIniFile.Create(Format(rsRemPluginsFile, [ExtractFilePath(ParamStr(0))]));
  FList          := TList.Create();
  FList.Capacity := 255;

  Res := FindFirst(FDirName + '*.alg', faAnyFile, SRec);
  while Res = 0 do begin
    if not FIniFile.ReadBool(rsRempluginsSection, SRec.Name, false) then
      AddPlugins(SRec.Name)
    else
      if(FIniFile.ReadDateTime(SRec.Name, rsDateTime, FileDate(FDirName + SRec.Name)) <> FileDate(FDirName + SRec.Name)) then
        AddPlugins(SRec.Name);

    Res := FindNext(SRec);
  end;
  FindClose(SRec);

  FList.Capacity := FList.Count;
end;

destructor TAlgoPluginsList.Destroy;
var I: Integer;
begin
  for I := 0 to Count - 1 do Items[I].Free;
  FList.Free;
  FIniFile.Free;
end;

function TAlgoPluginsList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TAlgoPluginsList.GetItems(Index: Integer): TAlgoPlugins;
begin
  result := TAlgoPlugins(FList[Index]);
end;

function TAlgoPluginsList.IndexOf(S: String): Integer;
var I: integer;
begin
  result := -1;
  for I := 0 to Count - 1 do begin
    if(Trim(Items[I].PluginsInformation.iName) = Trim(S)) then begin
      result := i;
      break;
    end;
  end;
end;

function TAlgoPluginsList.IndexOf(S: TInfosSerial): Integer;
var I: integer;
begin
  result := -1;
  for I := 0 to Count - 1 do begin
    if(Items[I].PluginsInformation.iSerial = S) then begin
      result := i;
      break;
    end;
  end;
end;

function TAlgoPluginsList.IndexOfBest(const St: TStream; const Count: integer): integer;
var I,    MinI   ,
    Size, MinSize,
    OldPos       : Integer;
begin
  MinI    := -1;
  MinSize := High(MinSize);

  OldPos  := St.Position;
  for I := 0 to Self.Count - 1 do begin
    St.Seek(OldPos, soFromBeginning);
    Items[I].OnProgression := OnProgression;
    Size := Items[I].PackTest(St, Count);
    if((Size > 0) and (Size < MinSize)) then begin
      MinSize := Size;
      MinI    := I;
    end;
  end;
  St.Seek(OldPos, soFromBeginning);
  result := MinI;
end;

procedure TAlgoPluginsList.RemovePlugins(Index: Integer);
var FileName: TFileName;
begin
  FileName := Items[Index].FileName;
  Items[Index].Free;
  FList.Delete(Index);

  FIniFile.WriteBool(rsRemPluginsSection, ExtractFileName(FileName), true);
  FIniFile.WriteDateTime(ExtractFileName(FileName), rsDateTime, FileDate(FileName));
end;

end.
